/*
 *  MPEG-1 Real Time Encoder
 *
 *  Copyright (C) 2000 Michael H. Schimek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: preview.c,v 1.2 2000/06/06 13:04:17 michael Exp $ */

#ifdef HAVE_LIBXV

#include <assert.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include <X11/extensions/Xv.h>
#include <X11/extensions/Xvlib.h>

#include "../log.h"

static Display		*display;
static int		screen;
static Window 		root_window, window;
static GC		gc;

static Visual 		*visual, *default_visual;

static unsigned int 	version, revision;
static unsigned int 	major_opcode;
static unsigned int 	event_base;
static unsigned int 	error_base;

// static Atom		atom_filter;

static XvPortID		port;
static XvAdaptorInfo 	*pAdaptor, *pAdaptors;
static XvFormat 	*pFormat;
static XvImageFormatValues *pImgFormats;
static int 		nAdaptors, nImgFormats;
// static int 		nFormats;

static XShmSegmentInfo	shminfo;
static XvImage		*ximage;
static XEvent		event;
// static int		CompletionType = -1;

extern int		width, height;

/*
 *  Use a buffer fifo (depth 3-4), presentation time stamps and a combined
 *  event/delay handler with a separate thread. Ref buffer layout must
 *  be 4:2:0 then we save a conversion and copy operation here.
 */

void
packed_preview(unsigned char *buffer, int mb_cols, int mb_rows)
{
	unsigned char *Y, *U, *V;
	int y_pitch, uv_pitch;
	int y_pitch8, uv_pitch8;
	int col, row, line;

	y_pitch8 = (y_pitch = ximage->pitches[0]) * 8;
	uv_pitch8 = (uv_pitch = ximage->pitches[1]) * 8;

	for (row = 0; row < mb_rows; row++) {
		Y = ximage->data + ximage->offsets[0] + y_pitch8 * row * 2;
		V = ximage->data + ximage->offsets[1] + uv_pitch8 * row;
		U = ximage->data + ximage->offsets[2] + uv_pitch8 * row;

		for (col = 0; col < mb_cols; col++) {
			for (line = 0; line < 8; line++) {
				memcpy(Y + 0, buffer + 0 * 64, 8); // Y0
				memcpy(Y + 8, buffer + 2 * 64, 8); // Y1
				memcpy(Y + y_pitch8 + 0, buffer + 1 * 64, 8); // Y2
				memcpy(Y + y_pitch8 + 8, buffer + 3 * 64, 8); // Y3
				memcpy(U, buffer + 4 * 64, 8); // Cb
				memcpy(V, buffer + 5 * 64, 8); // Cr

				buffer += 8;	// next 8 x 1 samples
				Y += y_pitch;	// next row
				U += uv_pitch;
				V += uv_pitch;
			}

			buffer += 5 * 64;	// next macroblock
			Y += 16 - y_pitch8;	// same strip, next column
			U += 8 - uv_pitch8;
			V += 8 - uv_pitch8;
		}
	}

	XSync(display, False);

	XvShmPutImage(display, port, window, gc, ximage,
		0, 0, width, height, // source coordinates
		0, 0, width, height, // destination
		False); // wait for completion

	// do XNextEvent(display, &event);
	// while (event.type != CompletionType);
}

#define YV12 0x32315659

void
preview_init(void)
{
	int i, j, k;
	char *title = "MP1E Preview";
	XVisualInfo *pVisInfo, VisInfoTmpl;
	XSetWindowAttributes win_attr;

	ASSERTX("open X11 display", display = XOpenDisplay(NULL));

	screen = XDefaultScreen(display);
        root_window = RootWindow(display, screen);

	if (Success != XvQueryExtension(display, &version, &revision,
		&major_opcode, &event_base, &error_base)) {
		FAIL("Xv extension not available");
	}

	if (version < 2 || revision < 2)
		FAIL("Xv version number %d.%d mismatch, "
			"should be 2.2 or later\n", version, revision);

	XvQueryAdaptors(display, root_window, &nAdaptors, &pAdaptors);

	if (nAdaptors == 0)
		FAIL("No Xv adaptors found");

	// atom_filter	= XInternAtom(display, "XV_FILTER", False);

	for (i = 0; i < nAdaptors; i++) {
		pAdaptor = pAdaptors + i;

		pFormat = pAdaptor->formats;

		for (j = 0; j < pAdaptor->num_ports; j++) {


			port = pAdaptor->base_id + j;

			printv(3, "Probing Xv adaptor #%d/%d, Port #%d/%d\n",
				i, nAdaptors, j, (int) pAdaptor->num_ports);

			// Let's see if this is an XvImage port...

			if (!(pImgFormats = XvListImageFormats(display, port, &nImgFormats)))
				continue;

			if (Success != XvGrabPort(display, port, CurrentTime))
				continue;

			// ...and supports YVU 4:2:0

			for (k = 0; k < nImgFormats; k++)
				if (pImgFormats[k].id == YV12)
					goto create;

			XvUngrabPort(display, port, CurrentTime);
		}
	}

	FAIL("This display does not support the XvImage extension,\n"
	    "the YVU 4:2:0 format, or all ports are occupied\n");

create:

	// Create an image buffer

	ximage = (XvImage *) XvShmCreateImage(display, port,
		YV12, NULL, (width + 63) & -64, height, &shminfo);
	/*
	 *  Width is padded to encourage cache line alignment of rows.
	 *  We clip to the actual size with XvPutShmImage. (And hope
	 *  the driver won't try something stupid...)
	 */

	printv(3, "ximage: %dx%d, %d planes\npitches: ",
		ximage->width, ximage->height, ximage->num_planes);
	for (i = 0; i < ximage->num_planes; i++)
		printv(3, "%d ", ximage->pitches[i]);
	printv(3, "\noffsets: ");
	for (i = 0; i < ximage->num_planes; i++)
		printv(3, "0x%08x=%d ", ximage->offsets[i], ximage->offsets[i]);
	printv(3, "\n");

	assert(ximage->num_planes == 3);
	assert(ximage->pitches[1] == ximage->pitches[2]);

	// Now we need a shared memory segment to bypass the X11 socket

	shminfo.shmid = shmget(IPC_PRIVATE, ximage->data_size, IPC_CREAT | 0777);
	shminfo.shmaddr = ximage->data = shmat(shminfo.shmid, 0, 0);
	shmctl(shminfo.shmid, IPC_RMID, 0); // remove when we terminate

	shminfo.readOnly = False;

	if (!XShmAttach(display, &shminfo))
		FAIL("XShmAttach failed");

	// CompletionType = XShmGetEventBase(display) + ShmCompletion;
	// redundant

	// This allocates a suitable colormap for 8 bit depth if needed

	default_visual = XDefaultVisual(display, screen);

	VisInfoTmpl.visualid = pFormat->visual_id;

	ASSERTX("find suitable X11 visual",
		pVisInfo = XGetVisualInfo(display,
			VisualIDMask, &VisInfoTmpl, &i));

	visual = pVisInfo->visual;

	// Create a simple window

	win_attr.event_mask = 
		StructureNotifyMask |
		KeyPressMask |
		ExposureMask |
		0;

	window = XCreateWindow(display, root_window,
		200, 200, width, height,
		0, pFormat->depth, InputOutput, visual,
		CWEventMask, &win_attr);

	ASSERTX("create a window", window);

	XSetStandardProperties(display, window,
		title, title, None, NULL, 0, NULL);

	XMapWindow(display, window);

	do XNextEvent(display, &event);
	while (event.type != MapNotify || event.xmap.event != window);

	gc = XCreateGC(display, window, 0, 0);

	// XShmDetach(display, &shminfo);
	// shmdt(shminfo.shmaddr);
	// XDestroyWindow(display, window);
}

#else

void
packed_preview(unsigned char *buffer, int mb_cols, int mb_rows)
{
}

void
preview_init(void)
{
}

#endif // HAVE_LIBXV
