/*
 *  MPEG-1 Real Time Encoder
 *
 *  Copyright (C) 1999-2000 Michael H. Schimek
 *
 *  Based on code written by Tom G. Lane
 *  and released to public domain 11/22/93.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: ieee.c,v 1.3 2000/05/01 13:52:41 michael Exp $ */

#include <limits.h>
#include "ieee.h"
#include "../mpeg.h"

static short random_block[8][8];	// Random source data
static short ref_fdct_block[8][8];	// Coefficients from reference FDCT
static short ref_idct_block[8][8];	// Coefficients from reference IDCT
static short test_block[8][8];		// Coefficients from test DCT

static int failed;

static double max_errors[64];
static double sum_errors[64];
static double sum2_errors[64];

/*
 *  IEEE 1180-1990 pseudo-random number generator
 */
static long randx = 1;
static double z = (double) 0x7fffffff;

static void
reset_rng(void)
{
	randx = 1;
}

static long
ieee_rand(long L, long H)
{
	long i,j;
	double x;

	randx = (randx * 1103515245) + 12345;
	i = randx & 0x7ffffffe;
	x = ((double) i) / z;
	x *= L + H + 1;
	j = x;

	return j - L;
}

void
ieee_randomize(short block[8][8], long minpix, long maxpix, long sign)
{
	int i;

	for (i = 0; i < 64; i++)
		block[0][i] = (short)(ieee_rand(-minpix, maxpix) * sign);
}

/*
 *  This routine creates patterns rarely appearing with ieee_randomize,
 *  but frequently causing overflows.
 */
void
rake_pattern(short block[8][8], long minpix, long maxpix, long sign)
{
	int i, j;
	int n;

	for (i = 0; i < 8; i++) {
		n = rand() & 1;
		for (j = 0; j < 8; j++)
			block[i][j] = (n ? minpix : maxpix) * sign;
	}

	if (rand() & 1)
		mirror(block);
}

void (* randomize)(short [8][8], long, long, long) = ieee_randomize;

/*
 *  IEEE rounding (round to nearest integer, halfway cases towards zero)
 */
int
ieee_round(double val)
{
	if (val < 0.0)
		return - floor(0.5 - val);
	else
		return floor(val + 0.5);
}

static int
saturate(int val, int min, int max)
{
	if (val < min)
		return min;
	else if (val > max)
		return max;
	else
		return val;
}

#undef sign
#define sign(v) saturate(v, -1, +1)

static void
saturate_block(short block[8][8], int min, int max)
{
	int i;

	for (i = 0; i < 64; i++)
		block[0][i] = saturate(block[0][i], min, max);
}

/*
 *  IEEE double-precision reference FDCT and IDCT
 */

double coslut[8][8];

static void init_coslut(void) __attribute__ ((constructor));

static void
init_coslut(void)
{
	int a, b;
	double tmp;

	for (a = 0; a < 8; a++)
		for (b = 0; b < 8; b++) {
			tmp = cos((double)((a + a + 1) * b) * (IEEE_PI / 16.0));

			if (b == 0)
				coslut[a][b] = tmp * 0.5 / sqrt(2.0);
			else
				coslut[a][b] = tmp * 0.5;
		}
}

void
ieee_ref_fdct(short block[8][8])
{
	int x, y, u, v;
	double tmp, tmp2;
	double res[8][8];

	for (v = 0; v < 8; v++)
		for (u = 0; u < 8; u++) {
			tmp = 0.0;
			for (y = 0; y < 8; y++) {
				tmp2 = 0.0;
				for (x = 0; x < 8; x++)
					tmp2 += (double) block[y][x] * coslut[x][u];
				tmp += coslut[y][v] * tmp2;
			}
			res[v][u] = tmp;
		}

	for (v = 0; v < 8; v++)
		for (u = 0; u < 8; u++)
			block[v][u] = (short) ieee_round(res[v][u]);
}

void
ieee_ref_idct(short block[8][8])
{
	int x, y, u, v;
	double tmp, tmp2;
	double res[8][8];

	for (y = 0; y < 8; y++)
		for (x = 0; x < 8; x++)	{
			tmp = 0.0;

			for (v = 0; v < 8; v++) {
				tmp2 = 0.0;

				for (u = 0; u < 8; u++)
					tmp2 += (double) block[v][u] * coslut[x][u];

				tmp += coslut[y][v] * tmp2;
			}

			res[y][x] = tmp;
		}

	for (v = 0; v < 8; v++)
		for (u = 0; u < 8; u++)
			block[v][u] = (short) ieee_round(res[v][u]);
}

/*
 *  Reference quantization (ISO 11172-2, 13818-2)
 */

void
mpeg1_intra_iquant(int quant_scale, short block[8][8])
{
	int i, val;

	block[0][0] = saturate(block[0][0] * 8, -2048, 2047);

	for (i = 1; i < 64; i++) {
		val = (2 * block[0][i] * default_intra_quant_matrix[0][i]
			* (quant_scale * 2)) / 32;

		if (!(val & 1))
			val -= sign(val);

		block[0][i] = saturate(val, -2048, 2047);
	}
}

void
mpeg1_inter_iquant(int quant_scale, short block[8][8])
{
	int i, val;

	for (i = 0; i < 64; i++) {
		val = ((2 * block[0][i] + sign(block[0][i]))
			* default_inter_quant_matrix[0][i] * (quant_scale * 2)) / 32;

		if (!(val & 1))
			val -= sign(val);

		block[0][i] = saturate(val, -2048, 2047);
	}
}

/* Assumed q_scale_type = 0 */

void
mpeg2_intra_iquant(int quant_scale, short block[8][8])
{
	int i, val, sum;

	block[0][0] = saturate(block[0][0] * 8, -2048, 2047);

	for (i = 1; i < 64; i++) {
		val = (2 * block[0][i] * default_intra_quant_matrix[0][i]
			* (quant_scale * 2)) / 32;

		sum += block[0][i] = saturate(val, -2048, 2047);
	}

	if (!(sum & 1))
		block[7][7] ^= 1;
}

void
mpeg2_inter_iquant(int quant_scale, short block[8][8])
{
	int i, val, sum;

	for (i = 0; i < 64; i++) {
		val = ((2 * block[0][i] + sign(block[0][i]))
			* default_inter_quant_matrix[0][i] * (quant_scale * 2)) / 32;

		sum += block[0][i] = saturate(val, -2048, 2047);
	}

	if (!(sum & 1))
		block[7][7] ^= 1;
}

void
mpeg_intra_quant(int quant_scale, short block[8][8])
{
	int i, val, div;

	block[0][0] = saturate((block[0][0] + 4 * sign(block[0][0])) / 8, -255, +255);

	for (i = 1; i < 64; i++) {
		val = block[0][i];
		div = 2 * (quant_scale * 2) * default_intra_quant_matrix[0][i];
		block[0][i] = saturate((32 * val + sign(val) * (div >> 1)) / div, -255, +255);
	}
}

void
mpeg_inter_quant(int quant_scale, short block[8][8])
{
	int i, val, div;

	for (i = 0; i < 64; i++) {
		val = block[0][i];
		div = 2 * (quant_scale * 2) * default_inter_quant_matrix[0][i];
		block[0][i] = saturate((32 * val) / div, -255, +255);
	}
}

/*
 *  Error logging and reporting
 */

static void
reset_errors(void)
{
	int i;

	for (i = 0; i < 64; i++)
		max_errors[i] =	sum_errors[i] = sum2_errors[i] = 0.0;
}

static void
report(char *name, long minpix, long maxpix, long sign, int iterations)
{
	double max, total;
	int i, j;

	printf("%s test conditions: -L = %ld, +H = %ld, sign = %ld, iterations = %ld\n\n",
		name, minpix, maxpix, sign, iterations);

	printf("Peak absolute values of errors:\n");

	max = 0.0;

	for (i = 0; i < 64; i++) {
		if (max < max_errors[i])
			max = max_errors[i];
		printf(" %4d", (int) max_errors[i]);
		if (i % 8 == 7)
			printf("\n");
	}

	printf("Worst peak error = %8.4f\n\n", max);
	
	printf("Mean square errors:\n");

	max = total = 0.0;

	for (i = 0; i < 64; i++) {
		double error = sum2_errors[i] / ((double) iterations);
		total += sum2_errors[i];
		if (max < error) max = error;
		printf(" %8.4f", error);
		if (i % 8 == 7)
			printf("\n");
	}

	printf("Worst PMSE = %.6f\n", max);

	total /= (double) (64 * iterations);

	printf("Overall MSE = %.6f\n\n", total);

	printf("Mean errors:\n");

	max = total = 0.0;

	for (i = 0; i < 64; i++) {
		double error = sum_errors[i] / ((double) iterations);
		total += sum_errors[i];
		printf(" %8.4f", error);
		if (error < 0.0) error = -error;
		if (max < error) max = error;
		if (i % 8 == 7)
			printf("\n");
	}

	printf("Worst mean error = %.6f\n", max);

	total /= (double) (64 * iterations);

	printf("Overall mean error = %.6f\n\n", total);

	printf("Testing 0 input giving 0 output:\n");

	for (i = 0, j = 0; i < 64; i++)
		if (test_block[0][i]) {
			printf("Position %d of DCT(0) = %d\n", i, test_block[i]);
			failed++;
			j++;
		}

	printf("%d elements of DCT(0) were not zero\n\n", j);
}

void
q_fdct_test(qdct_func *qfdct, qdct_func *quant, long minpix, long maxpix, long sign,
	int iterations, unsigned int quant_mask)
{
	int ii, jj, i, j;
	int count = 0;

	reset_errors();
	reset_rng();

	for (ii = 0; ii < iterations; ii++)
		for (jj = 1; jj < 32; jj++)
			if (quant_mask & (1 << jj)) {
				randomize(random_block, minpix, maxpix, sign);

				// Reference qFDCT

				copy(ref_fdct_block, random_block);
				ieee_ref_fdct(ref_fdct_block);
				quant(jj, ref_fdct_block);

				// Candidate qFDCT

				copy(test_block, random_block);
				qfdct(jj, test_block);

				// Accumulate errors

				for (i = 0; i < 64; i++) {
					int error = test_block[0][i] - ref_fdct_block[0][i];

					if (0 && error >= 8) {
						fprintf(stderr, "FATAL ii=%d q=%d error=%d at %d:\n", ii, jj, error, i);
						dump(random_block);
					}
					
					sum_errors[i] += error;
					sum2_errors[i] += error * error;

					error = abs(error);

					if (max_errors[i] < error)
						max_errors[i] = error;
				}

				if ((ii + jj) % 100 == 99) {
					// fprintf(stderr, ".");
					fprintf(stderr, "q_fdct_test %d%%\r", 100 * (ii + 1) / iterations);
					fflush(stderr);
				}
				
				count++;
			}

	clear(test_block);
	qfdct(1, test_block);

	fprintf(stderr, "\n\n");

	report("q-fdct", minpix, maxpix, sign, count);
}

void
q_idct_test(qdct_func *qidct, qdct_func *quant, qdct_func *iquant,
	long minpix, long maxpix, long sign, int iterations, unsigned int quant_mask)
{
	int ii, jj, i, j;
	int count = 0;

	reset_errors();
	reset_rng();

	for (ii = 0; ii < iterations; ii++)
		for (jj = 1; jj < 32; jj++)
			if (quant_mask & (1 << jj)) {
				randomize(random_block, minpix, maxpix, sign);

				// Reference qFDCT

				copy(ref_fdct_block, random_block);
				ieee_ref_fdct(ref_fdct_block);
				quant(jj, ref_fdct_block);

				// Reference qIDCT

				copy(ref_idct_block, ref_fdct_block);
				iquant(jj, ref_idct_block);
				ieee_ref_idct(ref_idct_block);

				// Candidate qIDCT

				copy(test_block, ref_fdct_block);
				qidct(jj, test_block);

				// Accumulate errors

				for (i = 0; i < 64; i++) {
					int error = test_block[0][i] - ref_idct_block[0][i];

					sum_errors[i] += error;
					sum2_errors[i] += error * error;
		    
					if (error < 0)
						error = -error;
	
					if (max_errors[i] < error)
						max_errors[i] = error;
				}

				if ((ii + jj) % 100 == 99) {
					// fprintf(stderr, ".");
					fprintf(stderr, "q_idct_test %d%%\r", 100 * (ii + 1) / iterations);
					fflush(stderr);
				}
				
				count++;
			}

	clear(test_block);
	qidct(1, test_block);

	fprintf(stderr, "\n\n");

	report("q-idct", minpix, maxpix, sign, count);
}

#define meets(val, lim) (((val) <= (lim)) ? "meets" : (failed++, "FAILS"))

static void
ieee_report(char *name, long minpix, long maxpix, long sign, int iterations)
{
	double max, total;
	int i, j;

	printf("IDCT: %s\nIEEE test conditions: -L = %ld, +H = %ld, sign = %ld, iterations = %ld\n\n",
		name, minpix, maxpix, sign, iterations);
		
	printf("Peak absolute values of errors:\n");

	max = 0.0;

	for (i = 0; i < 64; i++)
	{
		if (max < max_errors[i])
			max = max_errors[i];

		printf(" %4d", (int) max_errors[i]);

		if (i % 8 == 7)
			printf("\n");
	}

	printf("Worst peak error = %8.4f  (%s spec limit 1)\n\n", max, meets(fabs(max), 1.0));
	
	printf("Mean square errors:\n");

	max = total = 0.0;

	for (i = 0; i < 64; i++)
	{
		double error = sum2_errors[i] / ((double) iterations);

		total += sum2_errors[i];

		if (max < error) max = error;

		printf(" %8.4f", error);

		if (i % 8 == 7)
			printf("\n");
	}

	printf("Worst pmse = %.6f  (%s spec limit 0.06)\n", max, meets(fabs(max), 0.06));

	total /= (double) (64 * iterations);

	printf("Overall mse = %.6f  (%s spec limit 0.02)\n\n", total, meets(fabs(total), 0.02));
	
	printf("Mean errors:\n");

	max = total = 0.0;

	for (i = 0; i < 64; i++)
	{
		double error = sum_errors[i] / ((double) iterations);

		total += sum_errors[i];

		printf(" %8.4f", error);

		if (error < 0.0) error = -error;
		if (max < error) max = error;

		if (i % 8 == 7)
			printf("\n");
	}

	printf("Worst mean error = %.6f  (%s spec limit 0.015)\n", max, meets(fabs(max), 0.015));

	total /= (double) (64 * iterations);

	printf("Overall mean error = %.6f  (%s spec limit 0.0015)\n\n", total, meets(fabs(total), 0.0015));

	printf("Testing 0 input giving 0 output:\n");

	for (i = 0, j = 0; i < 64; i++)
		if (test_block[0][i]) {
			printf("Position %d of IDCT(0) = %d (FAILS)\n", i, test_block[i]);
			failed++;
			j++;
		}

	printf("%d elements of IDCT(0) were not zero\n\n", j);
}

void
ieee_idct_test(char *name, dct_func *idct, long minpix, long maxpix, long sign, int iterations)
{
	int ii, i, j;

	reset_errors();
	reset_rng();

	for (ii = 0; ii < iterations; ii++) {
		randomize(random_block, minpix, maxpix, sign);

		// Reference FDCT

		copy(ref_fdct_block, random_block);
		ieee_ref_fdct(ref_fdct_block);
		saturate_block(ref_fdct_block, -2048, 2047);

		// Reference IDCT

		copy(ref_idct_block, ref_fdct_block);
		ieee_ref_idct(ref_idct_block);
		saturate_block(ref_idct_block, -256, 255);

		// Candidate IDCT

		copy(test_block, ref_fdct_block);
		idct(test_block);
		saturate_block(test_block, -256, 255);

		// Accumulate errors

		for (i = 0; i < 64; i++) {
			int error = test_block[0][i] - ref_idct_block[0][i];

			sum_errors[i] += error;
			sum2_errors[i] += error * error;

			if (error < 0) error = -error;

			if (max_errors[i] < error)
				max_errors[i] = error;
		}

		if (ii % 100 == 99) {
			// fprintf(stderr, ".");
			fprintf(stderr, "idct_test %d%%\r", 100 * (ii + 1) / iterations);
			fflush(stderr);
		}
	}

	// Zero input test

	clear(test_block);
	idct(test_block);

	fprintf(stderr, "\n\n");

	ieee_report(name, minpix, maxpix, sign, iterations);
}

void
ieee_1180(char *name, dct_func *idct)
{
	printf("IEEE 1180-1990 IDCT test\n\n");

	failed = 0;
	randomize = ieee_randomize;

	ieee_idct_test(name, idct, -256, 255,  1, 10000);
	ieee_idct_test(name, idct,   -5,   5,  1, 10000);
	ieee_idct_test(name, idct, -300, 300,  1, 10000);

	ieee_idct_test(name, idct, -256, 255, -1, 10000);
	ieee_idct_test(name, idct,   -5,   5, -1, 10000);
	ieee_idct_test(name, idct, -300, 300, -1, 10000);

	printf("Test finished, %d failures\n\n", failed);
}

void
fdct_test(char *name, dct_func *fdct, long minpix, long maxpix, long sign, int iterations)
{
	int ii, i, j;

	reset_errors();
	reset_rng();

	for (ii = 0; ii < iterations; ii++) {
		randomize(random_block, minpix, maxpix, sign);

		// Reference FDCT

		copy(ref_fdct_block, random_block);
		ieee_ref_fdct(ref_fdct_block);
		saturate_block(ref_fdct_block, -2048, 2047);

		// Candidate FDCT

		copy(test_block, random_block);
		fdct(test_block);
		saturate_block(test_block, -2048, 2047);

		// Accumulate errors

		for (i = 0; i < 64; i++) {
			int error = test_block[0][i] - ref_fdct_block[0][i];

			sum_errors[i] += error;
			sum2_errors[i] += error * error;

			if (error < 0) error = -error;

			if (max_errors[i] < error)
				max_errors[i] = error;
		}

		if (ii % 100 == 99) {
			// fprintf(stderr, ".");
			fprintf(stderr, "fdct_test %d%%\r", 100 * (ii + 1) / iterations);
			fflush(stderr);
		}
	}

	// Zero input test

	clear(test_block);
	fdct(test_block);

	fprintf(stderr, "\n\n");

	report(name, minpix, maxpix, sign, iterations);
}
